package com.cirrial.digshark {
import com.cirrial.digshark.enemy.SeaPony;
import com.cirrial.digshark.enemy.Shell;
import com.cirrial.digshark.particles.DirtEmitter;
import com.cirrial.digshark.particles.DirtEmitter;
import com.cirrial.digshark.particles.ExplodeEmitter;
import com.cirrial.digshark.particles.FanEmitter;
import com.cirrial.digshark.particles.LaserEmitter;
import com.cirrial.digshark.widgets.Core;
import com.cirrial.digshark.widgets.DrillLaser;
import com.cirrial.digshark.widgets.Fan;
import com.cirrial.digshark.widgets.IActivatable;
import com.cirrial.digshark.widgets.IBiteable;
import com.cirrial.digshark.widgets.Nebula;
import com.cirrial.digshark.widgets.PressurePad;
import com.cirrial.digshark.widgets.Shark;
import com.cirrial.digshark.widgets.Teleporter;
import com.cirrial.util.C;

import org.flixel.*;
import org.osmf.layout.PaddingLayoutMetadata;

public class PlayState extends FlxState {

    public static var widgetGroup:FlxGroup;
    public static var nebulaGroup:FlxGroup;
    public static var pressurePadsGroup:FlxGroup;
    public static var sharksGroup:FlxGroup;
    public static var enemyGroup:FlxGroup;
    public static var particleGroup:FlxGroup;

    public static var player:Player;
    public static var core:Core;
    public static var map:FlxTilemap;

    public static var actionCooldown:Number = 0;

    public static var statusText:FlxText;
    public static var sharkText:FlxText;

    public static var forcedMoveMode:Boolean = false;
    public static var forceDirection:FlxPoint;

    public static var justHitPad:Boolean = false;
    public static var onPad:Boolean = false;

    public static var madeMove:Boolean = false;
    public static var savedShark:Boolean = false;
    public static var beatLevel:Boolean = false;
    public static var movesMade:uint = 0;

    public static var level:BaseLevel;

    override public function create():void {
        // initialise
        widgetGroup = new FlxGroup();
        nebulaGroup = new FlxGroup();
        pressurePadsGroup = new FlxGroup();
        sharksGroup = new FlxGroup();
        enemyGroup = new FlxGroup();
        particleGroup = new FlxGroup();

        forceDirection = new FlxPoint();

        forcedMoveMode = false;
        justHitPad = false;
        onPad = false;
        madeMove = false;
        savedShark = false;
        beatLevel = false;
        movesMade = 0;

        var levelClass:Class = Levels.levelData[Levels.current];
        level = new levelClass(true, onObjectAddedCallback);
        FlxG.camera.follow(player, FlxCamera.STYLE_TOPDOWN);
        FlxG.camera.setBounds(0, 0, BaseLevel.boundsMaxX, BaseLevel.boundsMaxY, true);
        FlxG.bgColor = level.bgColor;
        // level loaded, add stuff not defined in level editor

        attachPaths();

        core = new Core(0, 0);
        core.y = BaseLevel.boundsMaxY - core.height;
        add(core);

        add(particleGroup);

        statusText = new FlxText(0, 8, C.SCREEN_WIDTH);
        statusText.scrollFactor.x = 0.0;
        statusText.scrollFactor.y = 0.0;
        statusText.text = "Drill Energy: " + player.drillCharge;
        statusText.setFormat("gameFont", 30, 0xFFE0E0E0, "center", 0xFF000000);
        add(statusText);

        sharkText = new FlxText(0, 40, C.SCREEN_WIDTH);
        sharkText.scrollFactor.x = 0.0;
        sharkText.scrollFactor.y = 0.0;
        sharkText.text = "Shark Not Saved :C";
        sharkText.setFormat("gameFont", 20, 0xFFE04040, "center", 0xFF000000);
        add(sharkText);

        FlxG.flash(0xFF000000, 0.2);
    }

    override public function update():void {
        // update
        madeMove = false;

        // MOVEMENT
        if(actionCooldown <= 0) {
            if(!forcedMoveMode) {
                if(FlxG.keys.pressed(C.KEY_LEFT)) {
                    movePlayer(-1, 0);
                } else if(FlxG.keys.pressed(C.KEY_RIGHT)) {
                    movePlayer(1, 0);
                }
                if(FlxG.keys.pressed(C.KEY_UP)) {
                    movePlayer(0, -1);
                } else if(FlxG.keys.pressed(C.KEY_DOWN)) {
                    movePlayer(0, 1);
                }
            } else {
                switch(map.getTile(player.gridX + forceDirection.x, player.gridY + forceDirection.y)) {
                    case C.TILE_ICE:
                        player.move(forceDirection.x, forceDirection.y);
                        madeMove = true;
                        actionCooldown = C.ACTION_COOLDOWN/2;
                        break;
                    default:
                        forcedMoveMode = false;
                }
            }
        } else {
            actionCooldown -= C.EXPECTED_TIMESTEP;
        }

        // INTERACTION
        // bite/interact
        if(FlxG.keys.justPressed(C.KEY_BITE)) {
            FlxG.overlap(player, widgetGroup, onPlayerWidgetInteract);
            madeMove = true;
        }
        // overlaps for player
        onPad = false;
        FlxG.overlap(player, widgetGroup, onPlayerWidgetOverlap);
        FlxG.overlap(player, nebulaGroup, onPlayerNebulaOverlap);
        FlxG.overlap(sharksGroup, widgetGroup, onSharkWidgetOverlap);
        FlxG.overlap(player, enemyGroup, onPlayerEnemyOverlap);

        // pads
        if(!onPad) {
            for each(var pad:PressurePad in pressurePadsGroup.members) {
                pad.empty();
            }
        }

        // STEP OTHER ENTITIES
        if(madeMove) {
            movesMade++;
            var ent:Entity;
            for each(ent in widgetGroup.members) {
                ent.step();
            }
            for each(ent in enemyGroup.members) {
                ent.step();
            }
        }

        // TERMINATION
        // restart the level
        if(!player.alive || FlxG.keys.justPressed(C.KEY_RESTART)) {
            FlxG.fade(0xFF000000, 0.2, restartLevel);
        }
        // beat the level
        if(!beatLevel && player.gridY >= map.heightInTiles - 1) {
            winLevel();
        }

        // UI
        statusText.text = "Drill Energy: " + player.drillCharge;
        if(player.drillCharge == 0) {
            statusText.color = 0xFFD04040;
        } else {
            statusText.color = 0xFFC0C0C0;
        }


        Levels.framesTaken++;
        super.update();
    }

    public static function onPlayerWidgetOverlap(player:Player, widget:Entity):void {
        if(widget is PressurePad && !justHitPad) {
            justHitPad = true;
            onPad = true;
            widget.onCollide(player);
        } else {
            widget.onCollide(player);
        }
    }

    public static function onPlayerNebulaOverlap(player:Player, nebula:Nebula):void {
        if((player.gridX == nebula.gridX) && (player.gridY == nebula.gridY))
            player.back();
    }

    public static function onPlayerEnemyOverlap(player:Player, enemy:Entity):void {
        if(enemy is SeaPony) {
            if(((player.gridX == enemy.gridX) && (player.gridY == enemy.gridY)) || ((player.gridX == enemy.lastGridX) && (player.gridY == enemy.lastGridY)))
                if(player.gridX != player.lastGridX || player.gridY != player.lastGridY) {
                    player.back();
                } else {
                    player.shove(enemy);
                }
        } else if(enemy is Shell) {
            if((player.gridX == enemy.gridX) && (player.gridY == enemy.gridY)) {
                player.back();
            }
        }
    }

    public static function onSharkWidgetOverlap(shark:Shark, widget:Entity):void {
        if(widget is Teleporter && shark.gridX == widget.gridX && shark.gridY == widget.gridY) {
            FlxG.flash(0xFFC080F0, 0.2);
            sharkText.text = "Shark Saved! :D";
            sharkText.color = 0xFF40E040;
            savedShark = true;
            C.playSound(SoundLibrary.SndSaveShark, shark);
            shark.kill();
        }
    }

    public static function onPlayerWidgetInteract(player:Player, widget:Entity):void {
        widget.onBite();
    }

    public static function onObjectAddedCallback(obj:Object, group:FlxGroup, lvl:BaseLevel, scrollX:Number, scrollY:Number, properties:Array):Object {
        if(obj is Player) {
            player = obj as Player;
        } else if(obj is FlxTilemap) {
            map = obj as FlxTilemap;
        } else if(obj is Fan) {
            var fan:Fan = obj as Fan;
            fan.setProps(properties[0].value, properties[1].value, properties[2].value, properties[3].value);
            widgetGroup.add(fan);
        } else if(obj is DrillLaser) {
            var laser:DrillLaser = obj as DrillLaser;
            laser.setProps(properties[0].value, properties[1].value);
            widgetGroup.add(laser);
        } else if(obj is Nebula) {
            nebulaGroup.add(obj as Nebula);
        } else if(obj is PressurePad) {
            var pad:PressurePad = obj as PressurePad;
            pressurePadsGroup.add(pad);
            widgetGroup.add(pad);
        } else if(obj is Shark) {
            var shark:Shark = obj as Shark;
            sharksGroup.add(shark);
            widgetGroup.add(shark);
        } else if(obj is ObjectLink) {
            var link:ObjectLink = obj as ObjectLink;
            if(link.fromObject is IBiteable) {
                var bitable:IBiteable = link.fromObject as IBiteable;
                bitable.addActivatable(link.toObject as IActivatable);
            }
        } else if(obj is SeaPony || obj is Shell) {
            var enemy:Entity = obj as Entity;
            enemyGroup.add(enemy);
        } else if(obj is Entity) {
            widgetGroup.add(obj as Entity);
        }

        return obj;
    }

    public static function restartLevel():void {
        movesMade = 0;
        if(savedShark) Levels.sharksSaved[Levels.current] = false;
        FlxG.resetState();
    }

    public static function winLevel():void {
        C.playSound(SoundLibrary.SndWinLevel, player);
        beatLevel = true;
        Levels.movesMade += movesMade;
        if(savedShark) Levels.sharksSaved[Levels.current] = true;
        Levels.current++;
        if(Levels.current < Levels.levelData.length) {
            FlxG.fade(0xFFFFFFFF, 0.2, restartLevel);
        } else {
            FlxG.switchState(new ChooseState());
        }
    }

    public static function movePlayer(xDir:int, yDir:int, force:Boolean=false):void {
        if(actionCooldown <= 0) {
            switch(map.getTile(player.gridX + xDir, player.gridY + yDir)) {
                case C.TILE_BACKDROP:
                    if(force || (yDir >= 0 && !force)) {
                        player.move(xDir, yDir);
                    }
                    break;
                case C.TILE_DIRT:
                    if(player.drillCharge > 0) {
                        blowUpDirt(player.gridX + xDir, player.gridY + yDir);
                        player.drillCharge--;
                    } else {
                        break;
                    }
                case C.TILE_WATER:
                    player.move(xDir, yDir);
                case C.TILE_ICE:
                    forcedMoveMode = true;
                    forceDirection.x = xDir;
                    forceDirection.y = yDir;
            }
            actionCooldown = C.ACTION_COOLDOWN;
            madeMove = true;
        }
    }

    public static function attachPaths():void {
        for each (var pathData:PathData in level.paths) {
            if(pathData.childSprite) {
                var pony:SeaPony = pathData.childSprite as SeaPony;
                pony.setPathNodes(pathData.nodes);
            }
        }
    }

    public static function blowUpDirt(x:uint, y:uint):void {
        map.setTile(x, y, C.TILE_BACKDROP);
        C.playSoundAtPoint(SoundLibrary.SndDig, (x + 0.5)*C.TILE_SIZE, (y + 0.5)*C.TILE_SIZE);
        var dirtEmit:DirtEmitter = particleGroup.recycle(DirtEmitter) as DirtEmitter;
        dirtEmit.boom(x, y);
    }

    public static function makeDirt(x:uint, y:uint):void {
        map.setTile(x, y, C.TILE_DIRT);
        C.playSoundAtPoint(SoundLibrary.SndDirtLand, (x + 0.5)*C.TILE_SIZE, (y + 0.5)*C.TILE_SIZE);
        var dirtEmit:DirtEmitter = particleGroup.recycle(DirtEmitter) as DirtEmitter;
        dirtEmit.boom(x, y);
    }

    public static function explode(source:FlxSprite):void {
        var explodeEmit:ExplodeEmitter = particleGroup.recycle(ExplodeEmitter) as ExplodeEmitter;
        C.playSound(SoundLibrary.SndEnemyDie, source);
        explodeEmit.boom(source.x + source.width/2, source.y + source.height/2);
    }

    public static function laserZap(x:uint, y:uint):void {
        var laserEmit:LaserEmitter = particleGroup.recycle(LaserEmitter) as LaserEmitter;
        laserEmit.zap(x, y);
    }

    public static function getFanEmitter():FanEmitter {
        return particleGroup.recycle(FanEmitter) as FanEmitter;
    }
}
}